"""
Utility functions for running commands and shell scripts on a server without interactively logged in.
"""

from paramiko.client import SSHClient


class TeleShell(object):
    """This class is for running commands on servers."""

    def __init__(self, *, host, user, known_hosts_file, key_file):
        ssh = SSHClient()
        ssh.load_host_keys(filename=known_hosts_file)
        ssh.connect(hostname=host, username=user, key_filename=key_file)
        self._ssh = ssh

    def command(self, cmd):
        """
        Function to run single command.

        Args:
            cmd: Single command
        Returns:
            list: command result
            string: error message if command fails, None is succeeds
        """
        sshClient = self._ssh
        stdin, stdout, stderr = sshClient.exec_command(cmd)
        outputs = [each for each in stdout.readlines()]
        error = stderr.readlines()
        if len(error) != 0:
            error = error[0]
        else:
            error = None
        return outputs, error

    def commands(self, cmds):
        """
        Function to run multiple commands.

        Args:
            cmds: commands list
        Returns:
            list: commands result list
            string: error message is command fails, None is succeeds
        """
        cmd = '\n'.join(c for c in cmds)
        return self.command("bash -c '{str}'".format(str=cmd))

    def script(self, filename):
        """
        Function to run shell script file.

        Args:
            filename: name of script file
        Returns:
            list: commands result
            string: error message is command fails, None is succeeds
        """
        with open(filename, 'r') as f:
            comm = f.readlines()
        comms = ' '.join(each for each in comm)
        outputs, errors = self.command("bash -c '{str}'".format(str=comms))
        return outputs, errors

    def __call__(self, cmd):
        outputs, error = self.command(cmd)
        for line in outputs:
            print(line, end='')
        if error is not None:
            print(error)



'''
Use a 'bot' to post message to a Slack channel.
This is useful for error alerts and scheduled notifications.
'''

import json
import threading
import urllib.request

import arrow

# Reference:
# search for 'incoming webhooks for slack'

def post(channel_webhook_url: str, subject: str, text: str) -> None:
    dt = arrow.utcnow().to('US/Pacific').format('YYYY-MM-DD HH:mm:ss') + ' Pacific'
    json_data = json.dumps({
        'text': '--- {} ---\n{}\n{}'.format(subject, dt, text)
        }).encode('ascii')
    req = urllib.request.Request(
        channel_webhook_url, data=json_data, 
        headers={'Content-type': 'application/json'})
    thr = threading.Thread(target=urllib.request.urlopen, args=(req, ))
    thr.start()



import logging
import os
from typing import List, Tuple

import pymysql

from .sql import SQLClient

logger = logging.getLogger(__name__)


class MySQL(SQLClient):
    def __init__(self,
                 *,
                 user,
                 password,
                 database,
                 host,       # MySQL server url
                 port=3306,  # We use a non-standard port, so this is mandatory
                 **kwargs):
        super().__init__(
            conn_func=pymysql.connect,
            user=user,
            password=password,
            database=database,
            port=int(port),
            **kwargs)

    @property
    def closed(self):
        return self._conn.closed

    @property
    def connected(self):
        return not self.closed

    def reconnect(self):
        if self.closed:
            self._connect()

    def rollback(self):
        '''Call this after `write` but before `commit` to cancel the write.'''
        self._conn.rollback()

    def commit(self):
        '''Call this after `write`.'''
        self._conn.commit()



import pandas as pd


default = {k: pd.get_option(k) for k in
    ['display.max_rows', 'display.max_columns', 'display.max_colwidth',
    'display.width', 'display.float_format']}


def set_pandas_full_display(max_rows=256, max_columns=20, max_colwidth=200, display_width=600):
    pd.set_option('display.max_rows', max_rows)
    pd.set_option('display.max_columns', max_columns)
    pd.set_option('display.max_colwidth', max_colwidth)
    pd.set_option('display.width', display_width)
    pd.set_option('display.float_format', 
        lambda x: '{:.2f}'.format(x) if x >= 1. else '{:.2g}'.format(x))


def set_pandas_default_display():
    for k,v in default.items():
        pd.set_option(k, v)



from datetime import datetime
from pytz import timezone
from typing import List, Union, Tuple

import arrow


def tznow(tzname):
    """
    Get the current ``datetime`` in the specified timezone.

    Args:
        tzname: the ISO standard timezone name, such as ``Africa/Cairo``, ``America/Los_Angeles``.
    """
    return datetime.now(timezone(tzname))


def tzdate(tzname):
    """
    Get the current date part of ``datetime`` in the specified timezone.

    Args:
        tzname: the ISO standard timezone name, such as ``Africa/Cairo``, ``America/Los_Angeles``.
    """
    return datetime.now(timezone(tzname)).date()


def tznow_str(tzname):
    """
    Get the current ``datetime`` in the specified timezone,
    as a string formatted like '2016-09-03 18:22:03'.

    Args:
        tzname: the ISO standard timezone name, such as ``Africa/Cairo``, ``America/Los_Angeles``.
    """
    return tznow(tzname).strftime('%Y-%m-%d %H:%M:%S')


def tzdate_str(tzname):
    """
    Get the date part of current ``datetime`` in the specified timezone,
    as a string formatted like '2016-09-03'.

    Args:
        tzname: the ISO standard timezone name, such as ``Africa/Cairo``, ``America/Los_Angeles``.
    """
    return tzdate(tzname).strftime("%Y-%m-%d")


def shift_day(day: str, ndays: int) -> str:
    '''
    Given a date like '2018-09-23', shift it by (positive or negative) `ndays` days, e.g.

        shift_day('2018-09-23', -1)    # '2018-09-22'
        shift_day('2018-09-23', 10)    # '2018-10-03'

    This operation is ignorant of timezone.
    '''
    if ndays == 0:
        return day
    d = arrow.get(day, 'YYYY-MM-DD')
    return d.shift(days=ndays).format('YYYY-MM-DD')


def shift_hour(day: str, hour: str, nhours: int) -> Tuple[str, str]:
    '''
        shift_hour('2018-09-20', '08', 12)    # ('2018-09-20', '20')
        shift_hour('2018-09-20', '08', -24)   # ('2018-09-19', '08')
    '''
    if nhours == 0:
        return day, hour
    if len(hour) == 1:
        hour = '0' + hour
    dt = arrow.get(day + ' ' + hour, 'YYYY-MM-DD HH')
    dt = dt.shift(hours=nhours)
    return tuple(dt.format('YYYY-MM-DD HH').split())


class DateRange:
    def __init__(self,
                 start_date: str=None,
                 end_date: str=None,
                 n_days: int=None,
                 timezone: str='UTC'):
        '''
        `start_date` and `end_date` are either one of 'today', 'yesterday', 'tomorrow',
        or in the 'YYYY-MM-DD' format.

        `timezone`: other than the default 'UTC', other common values include 'US/Pacific', 'US/Eastern'.
        '''
        ff = 'YYYY-MM-DD'
        self._format = ff

        if timezone == 'UTC':
            today = arrow.utcnow().format('YYYY-MM-DD')
        else:
            today = arrow.now(timezone).format('YYYY-MM-DD')
        yesterday = shift_day(today, -1)
        tomorrow = shift_day(today, 1)
        common_dates = {'today': today, 'yesterday': yesterday, 'tomorrow': tomorrow}

        if start_date:
            start_date = common_dates.get(start_date, start_date)
        if end_date:
            end_date = common_dates.get(end_date, end_date)

        if start_date:
            if end_date:
                assert n_days is None
                assert start_date <= end_date
            else:
                assert n_days > 0
                end_date = shift_day(start_date, n_days - 1)
        else:
            assert end_date
            assert n_days > 0
            start_date = shift_day(end_date, -(n_days - 1))

        self._start_date = start_date
        self._end_date = end_date
        self._days = [start_date]
        while start_date < end_date:
            start_date = shift_day(start_date, 1)
            self._days.append(start_date)

    @property
    def days(self) -> List[str]:
        return self._days

    @property
    def n_days(self) -> int:
        return len(self._days)

    @property
    def first(self):
        return self._days[0]

    @property
    def last(self):
        return self._days[-1]



import logging
from pathlib import Path
import time
from typing import List

import boto3

# This module requires a directory `.aws/` containing credentials in the home directory,
# or environment variables `AWS_ACCESS_KEY_ID`, and `AWS_SECRET_ACCESS_KEY`.


logger = logging.getLogger(__name__)


def _get_client():
    return boto3.session.Session().client('s3')


def _has_key(s3_client, bucket: str, key: str) -> bool:
    response = s3_client.list_objects_v2(Bucket=bucket, Prefix=key)
    for obj in response.get('Contents', []):
        if obj['Key'] == key:
            return True
    return False


def _delete_key(s3_client, bucket: str, key: str) -> None:
    s3_client.delete_object(Bucket=bucket, Key=key)


def has_key(bucket: str, key: str) -> bool:
    return _has_key(_get_client(), bucket, key)


def delete_key(bucket: str, key: str) -> None:
    return _delete_key(_get_client(), bucket, key)


class Bucket:
    def __init__(self, bucket):
        for header in ('s3://', 's3n://'):
            if bucket.startswith(header):
                bucket = bucket[len(header) :]
                break
        if '/' in bucket:
            bucket = bucket[: bucket.find('/')]
        self._bucket = boto3.resource('s3').Bucket(bucket)

    @property
    def name(self):
        return self._bucket.name

    def _remove_bucket_key(self, key):
        for header in ('s3://', 's3n://'):
            if key.startswith(header):
                assert key.startswith(header + self.name + '/')
                key = key[(len(header) + len(self.name) + 1) :]
        return key

    def upload(self, local_file: str, s3_key: str) -> None:
        '''
        Upload a single file to S3.

        `local_file`: path to local file.
        `s3_key`: S3 'key'.
        
        Example: suppose current bucket is s3://my-org, with

        local_file: /home/zepu/work/data/xyz/memo.txt
        s3_key: mysurvey/memo
        --> remote file: s3://my-org/mysurvey/memo

        Existing file with the same name with be overwritten.
        '''
        local_file = Path(local_file)
        if not local_file.is_file():
            raise Exception('a file name is expected')
        data = open(local_file, 'rb')
        s3_key = self._remove_bucket_key(s3_key)
        self._bucket.put_object(Key=s3_key, Body=data)

    def upload_tree(self, local_path: str, s3_path: str,
                    pattern: str = '**/*') -> None:
        '''
        `local_path`: directory whose content will be uploaded.
            If `local_path` contains a trailing `/`, then no part of this path name
            becomes part of the remote name; otherwise, the final node in this path name
            becomes the leading segment of the remote name.
        `pattern`: 
            '*'    (everything directly under `local_path`),
            '**/*' (everything recursively under `local_path`),
            '*.py' (every Python module directly under `local_path`),
            '**/*.py' (every Python module recursively under `local_path`),
            etc.

        Example: suppose current bucket is s3://my-org, with
        
        local_path: /home/me/work/data/xyz, containing
            .../xyz/a.txt, 
            .../xyz/b.txt,
            ../xyz/zyx/aa.txt)
        s3_path: dataset1
        s3_name: '**/*'
        --> remote files: 
            s3://my-org/dataset1/xyz/a.txt
            s3://my-org/dataset1/xyz/b.txt
            s3://my-org/dataset1/xyz/zyx/aa.txt

        local_path: /home/me/work/data/xyz/ (note the trailing '/')
        --> remote files: 
            s3://my-org/dataset1/a.txt
            s3://my-org/dataset1/b.txt
            s3://my-org/dataset1/zyx/aa.txt
        '''
        with_root = not local_path.endswith('/')
        local_path = Path(local_path)
        if not local_path.is_dir():
            raise Exception('a directory name is expected')
        nodes = [v for v in local_path.glob(pattern) if v.is_file()]
        s3_path = self._remove_bucket_key(s3_path)
        for node in nodes:
            key = node.relative_to(local_path)
            if with_root:
                key = local_path.name / key
            key = s3_path / key
            self.upload(node, str(key))

    def download(self, s3_key: str, local_file: str = None) -> None:
        s3_key = self._remove_bucket_key(s3_key)
        if local_file is None:
            local_file = str(Path(s3_key).name)
        self._bucket.download_file(s3_key, local_file)

    def download_tree(self, s3_path: str, local_path: str = None) -> None:
        s3_path = self._remove_bucket_key(s3_path)
        raise NotImplementedError

    def ls(self, key, recursive: bool=False):
        # List object names directly or recursively named like `key*`.
        # If `key` is `abc/def/`,
        # then `abc/def/123/45` will return as `123/45`
        #
        # If `key` is `abc/def`,
        # then `abc/defgh/45` will return as `defgh/45`;
        # `abc/def/gh` will return as `/gh`.
        #
        # So if you know `key` is a `directory`, then it's a good idea to
        # include the trailing `/` in `key`.

        key = self._remove_bucket_key(key)
    
        z = self._bucket.objects.filter(Prefix=key)

        if key.endswith('/'):
            key_len = len(key)
        else:
            key_len = key.rfind('/') + 1

        if recursive:
            return (v.key[key_len :] for v in z)  
            # this is a generator, b/c there can be many, many elements
        else:
            keys = set()
            for v in z:
                vv = v.key[key_len :]
                idx = vv.find('/')
                if idx >= 0:
                    vv = vv[: idx]
                keys.add(vv)
            return sorted(list(keys))

    def has(self, key: str) -> bool:
        key = self._remove_bucket_key(key)

        if not hasattr(self, '_s3'):
            self._s3 = _get_client()
        return _has_key(self._s3, self._bucket.name, key)

    def delete(self, key: str) -> None:
        key = self._remove_bucket_key(key)

        if not hasattr(self, '_s3'):
            self._s3 = _get_client()
        _delete_key(self._s3, self._bucket.name, key)

    def delete_tree(self, s3_path: str) -> int:
        s3_path = self._remove_bucket_key(s3_path)
        
        n = 0
        while True:
            nn = self._delete_tree(s3_path)
            if nn == 0:
                break
            n = max(n, nn)
            time.sleep(0.5)
        return n

    def _delete_tree(self, s3_path: str) -> int:
        '''
        Return the number of objects deleted.

        After this operation, the 'folder' `s3_path` is also gone.
        
        TODO: this is not the fastest way to do it.
        '''
        assert s3_path.endswith('/')
        n = 0
        for k in self.ls(s3_path, recursive=True):
            kk = s3_path + k
            self.delete(kk)
            n += 1
        return n


def reduce_boto_logging():
    import boto3.s3.transfer
    for name in logging.Logger.manager.loggerDict.keys():
        if name.startswith('boto') or name.startswith('urllib3') or name.startswith('s3transfer'):
            logging.getLogger(name).setLevel(logging.ERROR)



import logging
from traceback import format_exc
import os
import random
from typing import List, Tuple, Union

from retrying import retry
import pyathena

from .hive import HiveTableMixin
from .sql import SQLClient
from ..s3 import reduce_boto_logging


import logging

logger = logging.getLogger(__name__)


TMP_DB = 'tmp'


def is_athena_error(e):
    if isinstance(e, pyathena.error.DatabaseError):
        logger.warning('error info before passed on to retrying')
        logger.warning(str(e))
        logger.warning(format_exc())
        return True
    return False


class Athena(SQLClient):
    def __init__(self, s3_result_dir: str = None) -> None:
        # `s3_result_dir` is where Athena query results are stored.
        # The default location is used if not specified.
        if s3_result_dir is None:
            s3_result_dir = 's3://aws-athena-query-results-{}-{}'.format(
                os.environ['AWS_ACCOUNT_ID'], os.environ['AWS_DEFAULT_REGION'])
        else:
            assert s3_result_dir.startswith('s3://')
        super().__init__(
            conn_func=pyathena.connect,
            s3_staging_dir=s3_result_dir,
            cursor_arraysize=1000,  # 1000 is Athena's upper limit
        )

    @retry(retry_on_exception=is_athena_error, 
            wait_exponential_multiplier=30000, # 30 seconds
            wait_exponential_max=120000,  # 2 minutes
            stop_max_attempt_number=7)
    def read(self, *args, **kwargs):
        return super().read(*args, **kwargs)

    @retry(retry_on_exception=is_athena_error, 
            wait_exponential_multiplier=30000, # 30 seconds
            wait_exponential_max=120000,  # 2 minutes
            stop_max_attempt_number=7)
    def write(self, *args, **kwargs):
        return super().write(*args, **kwargs)


def reduce_athena_logging():
    import pyathena.common
    for name in logging.Logger.manager.loggerDict.keys():
        if name.startswith('pyathena'):
            logging.getLogger(name).setLevel(logging.ERROR)

    reduce_boto_logging()


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_attempt_number=7)
def get_athena():
    return Athena()


class AthenaTable(HiveTableMixin):
    def __init__(self, location: str, **kwargs) -> None:
        assert location.startswith('s3://')
        super().__init__(location=location, **kwargs)

    def insert_overwrite_partition(self, engine: Athena, sql: str, *partition_values, purge_data: bool=True):
        '''
        `purge_data`: if `True`, purge existing data in the partion pointed to by `partition_values`.
            If `False`, do not purge; instead, assume the operation will write into sub-directories that
            are already purged, and there may be other subdirectories that should be left alone.

        See https://docs.aws.amazon.com/athena/latest/ug/create-table-as.html
        '''
        tmp_tb = 'tmp' + str(random.random()).replace('.', '').replace('-', '')
        tmp_tb = f'{TMP_DB}.{tmp_tb}'

        if len(partition_values) < len(self.partitions):
            parts = ', '.join([f"'{k}'" for k,v in self.partitions[len(partition_values): ]])
            parts = f'partitioned_by = ARRAY[{parts}],'
            # Be sure to verify that the last columns in `sql` match these partition fields.
        else:
            parts = ''

        ppath = self.partition_location(*partition_values)

        assert self.stored_as.lower() in ('orc', 'parquet')

        sql = f'''
            CREATE TABLE {tmp_tb}
            WITH (
                external_location = '{ppath}',
                format = '{self.stored_as}',
                {parts}
                {self.stored_as.lower()}_compression = '{self.compression}'
            )
            AS
            {sql}
        '''

        engine.write(f'DROP TABLE IF EXISTS {tmp_tb}')
        if purge_data:
            self.purge_data(*partition_values)
        logger.debug('\n' + sql)
        engine.write(sql)
        self.update_partitions(engine, *partition_values)
        engine.write(f'DROP TABLE IF EXISTS {tmp_tb}')

    @classmethod
    def from_hive_table(cls, table: 'HiveTable', db_name: str, tb_name: str=None) -> 'AthenaTable':
        '''
        Use case of this method:

            If we have defined and manipulated an _external_ table in Hive,
            then the data is in S3 but meta data is in Hadoop.
            This table is not available to Athena because the meta data is not in S3.
            The current method takes the existing Hive (external) table definition
            and creates an AthenaTable object, so that the table's meta data is in sync
            with the Hive table. After calling `create` on the AthenaTable object,
            the table is present in the S3 meta data store, so that the table can be used
            by Athena.

            The table meta data in Hadoop and in S3 are independent of each other.
            Therefore if the table's data is modified by either side of Hive and Athena,
            the other side needs to update partitions.
        '''
        assert table.s3external
        location = table.location
        if location.startswith('s3n://'):
            location = 's3://' + location[len('s3n://') :]
        return cls(
            db_name=db_name,
            tb_name=tb_name or table.tb_name,
            columns=table.columns,
            partitions=table.partitions,
            stored_as=table.stored_as,
            field_delimiter=table.field_delimiter,
            compression=table.compression,
            location=location)


import logging
from typing import List, Tuple

import psycopg2

from .sql import SQLClient

logger = logging.getLogger(__name__)


class Postgres(SQLClient):
    def __init__(self, *, host, user, password, dbname, port=5432, **kwargs):
        super().__init__(
            conn_func=psycopg2.connect,
            host=host,
            user=user,
            password=password,
            dbname=dbname,
            port=int(port),
            **kwargs)

    @property
    def closed(self):
        return self._conn.closed

    @property
    def connected(self):
        return not self.closed

    def reconnect(self):
        if self.closed:
            self._connect()

    def rollback(self):
        '''Call this after `write` but before `commit` to cancel the write.'''
        self._conn.rollback()

    def commit(self):
        '''Call this after `write`.'''
        self._conn.commit()

    def get_databases(self) -> List[str]:
        sql = "SELECT datname FROM pg_database WHERE datistemplate = false"
        headers, rows = self.read(sql).fetchall()
        return [v[0] for v in rows]

    def get_tables(self) -> List[str]:
        sql = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public'"
        rows = self.read(sql).fetchall()
        return [v[0] for v in rows]

    def has_table(self, tb_name):
        # A 'lazy' solution
        # return tb_name in self.get_tables()

        sql = "SELECT exists(SELECT relname FROM pg_class WHERE relname = '{}')".format(
            tb_name)
        return self.read(sql).fetchall()

    def get_table_schema(
            self, tb_name) -> Tuple[List[str], List[Tuple[str, str, str]]]:
        '''
        Returns: tuple with
            ['column_name', 'data_type', 'is_nullable'],
            [[column_name, data_type, is_nullable] for each field]
        '''
        sql = """
        SELECT column_name, data_type, is_nullable
        FROM information_schema.columns
        WHERE table_name = '{tb_name}'
        """.format(tb_name=tb_name)
        return self.read(sql).fetchall()

    def get_table_columns(self, tb_name) -> List[str]:
        sql = "SELECT column_name from information_schema.columns WHERE table_name = '{}'".format(
            tb_name)
        rows = self.read(sql).fetchall()
        return [v[0] for v in rows]


import logging
import textwrap
from typing import Union, Sequence, List, Tuple, Callable, Optional

import sqlparse
import pandas as pd

logger = logging.getLogger(__name__)


def cleanse_sql(x):
    return x.strip().strip('\n').strip().strip(';').strip()


def split_sql(sql: str) -> Sequence[str]:
    """
    Split a sequence of SQL statements into individual statements.

    The input is a block of text formatted like ::

        CREATE TABLE ...
        ;
        INSERT INTO ...
        SELECT ...
        ;
        DROP TABLE ...


    ``;\\n`` is treated as the separator between SQL statements.

    For each single SQL statement in the resultant list,
    trailing ';', line break, and spaces are removed;
    leading ';' and line break are removed;
    leading spaces are not removed b/c that may be part of intentional formatting
    for alignment.
    """
    sql.replace('\n\n', '\n')
    z = [cleanse_sql(x) for x in cleanse_sql(sql).split(';\n') if x]
    return [x for x in z if x]


def pretty_sql(sql: str, lpad: str = '        ', keyword_case=None) -> str:
    """
    Convert a SQL statement(s) into readable format (nicely line-split, indented, and aligned)
    suitable for printing and inspection.
    """
    assert keyword_case in (None, 'upper', 'lower')
    sql = sqlparse.format(
        cleanse_sql(sql),
        reindent=True,
        keyword_case=keyword_case,
        strip_whitespace=True,
    )
    sql = sql.lstrip('\n').rstrip(' \n').replace('\n\n\n', '\n\n')
    if lpad:
        sql = textwrap.indent(sql, lpad)
    return sql


class SQLReader:
    '''
    SQL client for a read-only engine.
    '''

    def __init__(self,
                 *,
                 conn_func: Callable,
                 cursor_args: dict = None,
                 cursor_arraysize: int = 10000,
                 **conn_args) -> None:
        """
        Args:
            conn_func: function 'connect' of the appropriate class.
            cursor_args: keyword args to 'connection.cursor()'.
                In most cases this should be left as `None`.
            **conn_args: connection parameters, typically include
                'port', 'user', 'password', and so on.
                See specific subclasses for specifics.
        """
        if 'port' in conn_args:
            conn_args['port'] = int(conn_args['port'])
        self._conn_func = conn_func
        self._conn_args = conn_args
        if cursor_args is None:
            cursor_args = {}
        self._cursor_args = cursor_args
        self._cursor_arraysize = cursor_arraysize

        self._connect()

    @property
    def user(self) -> Optional[str]:
        return self._conn_args.get('user')

    def _create_cursor(self):
        try:
            self._cursor = self._conn.cursor(**self._cursor_args)
            self._cursor.arraysize = self._cursor_arraysize
        except:
            logger.error(
                'Failed to create cursor on connection with arguments %s',
                str(self._cursor_args))
            raise

    def _connect(self) -> None:
        try:
            self._conn = self._conn_func(**self._conn_args)
        except:
            logger.error('Failed to connect to server with arguments %s' % str(
                self._conn_args))
            raise
        self._create_cursor()

    def close(self) -> None:
        try:
            self._cursor.close()
            self._conn.close()
        except:
            pass

    def __del__(self) -> None:
        self.close()

    def _execute(self, sql: str, **kwargs) -> None:
        '''Execute a single SQL statement.'''
        assert isinstance(sql, str)
        try:
            self._cursor.execute(sql, **kwargs)
        except:
            logger.error('Failed to execute:\n%s\nwith configuration: %s', sql, kwargs)
            raise

    def read(self, sql: str, **kwargs):
        self._execute(sql, **kwargs)
        return self

    def iterrows(self):
        """
        Iterator over rows in result after calling ``read``.

        Before ``_execute`` is ever run, using the iterator is an error.
        """
        return self._cursor

    def iterbatches(self, size=None):
        """
        This method is called after ``read`` to iter over results, one batch at a time.
        """
        while True:
            rows = self.fetchmany(size)
            if rows:
                yield rows
            else:
                break

    @property
    def headers(self) -> List[str]:
        """
        Return column headers after calling ``read``.

        This can be used to augment the returns of `fetchone`, `fetchmany`, `fetchall`,
        which return values only, i.e. they do not return column headers.
        """
        return [x[0] for x in self._cursor.description]

    def fetchone(self) -> Union[Sequence[str], None]:
        """
        Fetch the next row of a query result set, returning a single sequence, 
        or None when no more data is available.

        An Error (or subclass) exception is raised if the previous call to 
        ``read`` did not produce any result set or no call to ``read`` was issued yet.
        """
        return self._cursor.fetchone()

    def fetchone_pandas(self, simplify=False):
        row = self.fetchone()
        if not row:
            return pd.DataFrame(columns=self.headers)
        if simplify and len(row) == 1:
            return row[0]

        return pd.DataFrame.from_records([row], columns=self.headers)

    def fetchmany(self, n=None) -> Sequence[Sequence[str]]:
        """
        Fetch the next set of rows of a query result, returning a sequence of sequences
        (e.g. a list of tuples). 
        An empty sequence is returned when no more rows are available.

        An Error (or subclass) exception is raised if the previous call to 
        ``read`` did not produce any result set or no call to ``read`` was issued yet.
        """
        return self._cursor.fetchmany(n or self._cursor.arraysize)

    def fetchmany_pandas(self, n=None, simplify=False):
        rows = self.fetchmany(n)
        if not rows:
            return pd.DataFrame(columns=self.headers)
        if simplify and len(rows) == 1 and len(rows[0]) == 1:
            return rows[0][0]

        return pd.DataFrame.from_records(rows, columns=self.headers)

    def fetchall(self) -> Sequence[Sequence[str]]:
        """
        Fetch all (remaining) rows of a query result, returning them as a sequence of sequences
        (e.g. a tuple of tuples). 
        Note that the cursor's arraysize attribute can affect the performance of this operation.

        An Error (or subclass) exception is raised if the previous call to 
        ``read`` did not produce any result set or no call to ``read`` was issued yet.
        """
        return self._cursor.fetchall()

    def fetchall_pandas(self, simplify=False):
        """
        This method is called after ``read`` to fetch the results as a ``pandas.DataFrame``.
         
        `simplify`: if `True`, then if the resultant ``pandas.DataFrame``
            contains only one element, the single element is returned as a scalar
            (as opposed to a ``1 X 1 pandas.DataFrame``).

        Warning: do not use this if the result contains a large number of rows.
        """
        rows = self.fetchall()
        if not rows:
            return pd.DataFrame(columns=self.headers)
        if simplify and len(rows) == 1 and len(rows[0]) == 1:
            return rows[0][0]

        return pd.DataFrame.from_records(list(rows), columns=self.headers)


class SQLClient(SQLReader):
    '''
    SQL client for both read and write.
    '''

    def write(self, sql: Union[str, List[str]], **kwargs):
        if isinstance(sql, str):
            sql = split_sql(sql)
        for s in sql:
            self._execute(s, **kwargs)
        return self



import collections


def flatten(x):
    if x == []:
        return x
    if isinstance(x[0], list):
        return flatten(x[0]) + flatten(x[1:])
    return x[:1] + flatten(x[1:])


def ordered_dedupe(x):
    assert isinstance(x, list)
    return list(collections.OrderedDict.fromkeys(x))



from datetime import datetime, timedelta
import functools
import inspect
import json
import logging
import os
from pathlib import Path
from traceback import format_exc
from typing import Union, Optional

import arrow

from .profile import Timer
from .slack import post as post_to_slack


logger = logging.getLogger(__name__)


def notify_ok(subject: str, msg: str):
    # Need to a way to get the webhook URL of the slack channel, e.g.
    # via environment variables.
    url = 'abc' # for example
    post_to_slack(url, subject, msg)


def notify_error(subject: str, msg: str):
    # Need to a way to get the webhook URL of the slack channel, e.g.
    # via environment variables.
    url = 'def' # for example
    post_to_slack(url, subject, msg)


def log_notify(filename, status, info):
    dt = arrow.utcnow().format('YYYY-MM-DD HH:mm:ss.SSSSSS') + ' UTC'
    info = info.replace('\n', ' ')
    open(filename, 'w').write(status + '\n' + dt + '\n' + info)


def should_send_alert(status: str, ff: Path, silent_seconds: Union[float, int],
                      ok_silent_hours: Union[float, int], info: str) -> bool:
    if not ff.exists():
        return True

    info = info.replace('\n', ' ')

    try:
        old_status, old_dt, old_info = open(ff).read().split('\n')
        old_date, old_time, *_ = old_dt.split()
    except Exception as e:
        return True

    if old_status != status:
        return True

    if old_info != info:
        return True

    t0 = arrow.get(old_date + ' ' + old_time)
    t1 = arrow.utcnow()
    lapse = (t1 - t0).total_seconds()

    if lapse < float(silent_seconds):
        return False

    if status != 'OK':
        return True

    if lapse > float(ok_silent_hours) * 3600.:
        return True

    return False


def notify(exception_classes: Exception = None,
           debug: bool = False,
           with_args: bool = True,
           silent_seconds: Optional[Union[float, int]] = None,
           ok_silent_hours: Optional[Union[float, int]] = None):
    '''
    A decorator for writing a status file for a function for notification purposes.

    In addition, `slack_channel`, `silent_seconds`, `ok_silent_hours`, in combination with
    the current and previous statuses, determine whether to send alert to Slack.
    See `should_send_alert` for details.

    Args:
        exception_classes: exception class object, or tuple or list of multiple classes,
            to be captured; if `None`, all exceptions will be captured.

        silent_seconds: if new status is identical to the previous one and the previous
            status was written within the last `silent_seconds` seconds, do not send alert.

        ok_silent_hours: if new and previous statuses are both 'OK' and the previous status
            was written within the last `ok_silent_hours` hours, do not send alert.

        debug: if either `silent_seconds` or `ok_silent_hours` is `None`, their values are
            determined according whether `debug` is `True`.

    The status file is located in the directory specified by the environ variable `NOTIFYDIR`.
    The file name is constructed by the package/module of the decorated function as well as the function's name.
    For example, if function `testit` in package/module `my.models.regression.linear` is being decorated,
    then the status file is `my.models.regression.linear.testit` in the notify directory.
    
    If the decorated function is in a launching script (hence its `module` is named `__main__`),
    then the full path of the script is used to construct the status file's name.
    For example, if function `testthat` in script `/home/docker-user/work/scripts/do_this.py` is being decorated,
    then the status file is `home.docker-user.work.scripts.do_this.py.testthat` in the notify directory.
    
    This decorator writes 'OK' in the status file if the decorated function returns successfully.
    If the decorated function raises any exception, `CRITICAL` is written along with some additional info.
    
    This decorator does not write logs. If you wish to log the exception, you must handle that separately.
    If you handle the exception within the function, make sure you re-`raise` the exception so that this decorator
    can capture it (if you want it to).
    
    Usually you only need to decorate the top-level function in a pipeline's launching script, like this:
        
        # launcher.py

        from zpz.notify import notify

        @notify()
        def main():
            # do things that could raise exceptions
            # ...

        if __name__ == '__main__':
            main()

    You want to use this decorator at more refined places only when you want to handle a certain exception
    and then continue the program, but also want to notify about that exception, like this:

        # module `abc.py` in package `proj1.component2`

        from zpz.notify import notify

        class MySpecialError(Exception):
            pass

        @notify(MySpecialError)
        def func1():
            # do things that could raise MySpecialError
            # ...
            # result = ...
            if result is None:
                raise MySpecialError('omg!')

            #...
            #...

        def func2():
            try:
                func1()
                #...
            except MySpecialError as e:
                logger.info('MySpecialError has occurred!')
                return 3    # let program continue to run
    '''
    if not exception_classes:
        exception_classes = (Exception, )
    elif issubclass(exception_classes, Exception):
        exception_classes = (exception_classes, )
    else:
        if isinstance(exception_classes, list):
            exception_classes = tuple(exception_classes)
        assert isinstance(exception_classes, tuple)
        assert all(issubclass(v, Exception) for v in exception_classes)

    if silent_seconds is None:
        silent_seconds = 1.0

    if ok_silent_hours is None:
        if debug:
            ok_silent_hours = 1. / 60.  # 1 minute
        else:
            ok_silent_hours = 23.90  # 1 day

    def decorator(func):
        module = func.__module__
        if module == '__main__':
            module = str(Path.cwd() / inspect.getsourcefile(func))
        decloc = module + ' :: ' + func.__name__
        fname = module.strip('/').replace('/', '.') + '.' + func.__name__
        fdir = os.environ['NOTIFYDIR']
        Path(fdir).mkdir(parents=True, exist_ok=True)

        notifile = Path(fdir) / fname

        @functools.wraps(func)
        def decorated(*args, **kwargs):
            mytimer = Timer().start()
            if with_args:
                args_msg = f'args: {args}\nkwargs: {kwargs}\n'
            else:
                args_msg = ''
            status_msg = decloc + args_msg
            try:
                z = func(*args, **kwargs)
                status = 'OK'
                msg = '{}\nThe function took {} to finish.\n{}'.format(
                    decloc, timedelta(seconds=mytimer.stop().seconds), args_msg)
                if should_send_alert(status, notifile, silent_seconds,
                                     ok_silent_hours, status_msg):
                    notify_ok(status, msg)
                log_notify(notifile, status, status_msg)
                return z
            except exception_classes as e:
                status = 'ERROR'
                msg = '{}\n\n{}\n{}'.format(decloc, format_exc(), args_msg)
                if should_send_alert(status, notifile, silent_seconds,
                                     ok_silent_hours, status_msg):
                    notify_error(status, msg)
                log_notify(notifile, status, status_msg)
                raise
            except:
                status = 'OK'
                msg = '{}\nThe function took {} to finish.\n{}'.format(
                        decloc, timedelta(seconds=mytimer.stop().seconds), args_msg)
                if should_send_alert(status, notifile, silent_seconds,
                                     ok_silent_hours, status_msg):
                    notify_ok(status, msg)
                log_notify(notifile, status, status_msg)
                raise

        return decorated

    return decorator




import io

from sklearn.externals import joblib


def dump_bytes(x, compress: int = 9) -> bytes:
    """
    Serialize Python object (e.g. fitted model) `x` into a binary blob.
    """
    o = io.BytesIO()
    joblib.dump(x, o, compress=compress)
    return o.getvalue()


def load_bytes(b: bytes):
    """
    Inverse of `dump_bytes`.
    """
    return joblib.load(io.BytesIO(b))


def dump_file(x, filename: str, overwrite: bool = False,
              compress: int = 9) -> None:
    """
    Persist Python object (e.g. fitted model) `x` into disk file `filename`.
    """
    with open(filename, 'wb' if overwrite else 'xb') as f:
        f.write(dump_bytes(x, compress=compress))


def load_file(filename: str):
    """
    Inverse of `dump`.
    """
    with open(filename, 'rb') as f:
        z = f.read()
    return load_bytes(z)



import io
import json
from typing import Union
import zlib

import avro.schema
import avro.datafile
import avro.io
import numpy


class BinaryEncoder(avro.io.BinaryEncoder):
    def write_float(self, datum):
        # A `numpy.float32` will call this function and be stored as
        # 4-byte `float` of `C`.
        assert type(datum) is numpy.float32
        super().write_float(datum)

    def write_double(self, datum):
        assert type(datum) is float or type(datum) is numpy.float64
        super().write_double(datum)

    def write_int(self, datum):
        # Although Python `int` has unlimited bits, we'll treat it as `int` of `C`, i.e. 4 bytes.
        # If you know the numbers can be large, use `numpy.int64`, which will be treated as `long` of `C`.
        # This `int` is the common `int32`.
        #
        # Python `int` and `numpy.int32` will call this function and be stored
        # as 4-byte `int` of `C`.
        assert type(datum) is int or type(datum) is numpy.int32
        assert -2147483648 < datum < 2147483647
        datum = (datum << 1) ^ (datum >> 31)
        while (datum & ~0x7f) != 0:
            self.WriteByte((datum & 0x7f) | 0x80)
            datum >>= 7
        self.WriteByte(datum)

    def write_long(self, datum):
        super().write_long(datum)


class DatumWriter(avro.io.DatumWriter):
    def write(self, datum, encoder):
        # skip schema validation
        self.write_data(self.writer_schema, datum, encoder)

    def write_fixed(self, writer_schema, datum, encoder):
        if not isinstance(datum, bytes):
            datum = datum.tobytes()  # works if `datum` is a Numpy object
        super().write_fixed(writer_schema, datum, encoder)


class DataFileWriter(avro.datafile.DataFileWriter):
    def __init__(self, writer, writer_schema=None, codec='null'):
        if isinstance(writer_schema, dict):
            writer_schema = json.dumps(writer_schema)
        if isinstance(writer_schema, str):
            writer_schema = avro.schema.Parse(writer_schema)
        datum_writer = DatumWriter()
        super().__init__(writer, datum_writer, writer_schema, codec)
        self._encoder = BinaryEncoder(writer)
        self._buffer_encoder = BinaryEncoder(self._buffer_writer)


class BinaryDecoder(avro.io.BinaryDecoder):
    # `read_double` will read into Python `float`.

    def read_float(self):
        return numpy.float32(super().read_float())

    # `read_int` and `read_long` will both read into Python `int`.
    # The resultant number will then be put into `numpy` if
    # the schema specifies `pytype` as `numpy`.


class DatumReader(avro.io.DatumReader):
    def read_data(self, writer_schema, reader_schema, decoder):
        z = super().read_data(writer_schema, reader_schema, decoder)
        if writer_schema.props.get('pytype', None) == 'numpy':
            if writer_schema.type == 'long':
                z = numpy.int64(z)
            elif writer_schema.type == 'int':
                z = numpy.int32(z)
            elif writer_schema.type == 'double':
                z = numpy.float64(z)
            elif writer_schema.type == 'float':
                z = numpy.float32(z)
            elif writer_schema.type == 'array':
                z = numpy.array(z)
            elif writer_schema.type == 'fixed':
                assert type(z) is bytes
                dtype = numpy.dtype(writer_schema.props['logical_type'])
                z = numpy.frombuffer(z, dtype)[0]
        return z


class DataFileReader(avro.datafile.DataFileReader):
    def __init__(self, reader):
        # Same as superclass, except for replacing `avro.io.BinaryDecoder`
        # by `BinaryDecoder` (the custom version above).

        self._reader = reader
        self._raw_decoder = BinaryDecoder(reader)
        self._datum_decoder = None  # Maybe reset at every block.
        self._datum_reader = DatumReader()

        # read the header: magic, meta, sync
        self._read_header()

        # ensure codec is valid
        avro_codec_raw = self.GetMeta('avro.codec')
        if avro_codec_raw is None:
            self.codec = "null"
        else:
            self.codec = avro_codec_raw.decode('utf-8')
        if self.codec not in avro.datafile.VALID_CODECS:
            raise avro.datafile.DataFileException(
                'Unknown codec: %s.' % self.codec)

        self._file_length = self._GetInputFileLength()

        # get ready to read
        self._block_count = 0
        self.datum_reader.writer_schema = (avro.schema.Parse(
            self.GetMeta(avro.datafile.SCHEMA_KEY).decode('utf-8')))

    def _read_block_header(self):
        # Replace `avro_io.BinaryDecoder` in original implementation
        # by `BinaryDecoder` (our custom version above).
        # No other changes.

        self._block_count = self.raw_decoder.read_long()
        if self.codec == "null":
            # Skip a long; we don't need to use the length.
            self.raw_decoder.skip_long()
            self._datum_decoder = self._raw_decoder
        elif self.codec == 'deflate':
            # Compressed data is stored as (length, data), which
            # corresponds to how the "bytes" type is encoded.
            data = self.raw_decoder.read_bytes()
            # -15 is the log of the window size; negative indicates
            # "raw" (no zlib headers) decompression.  See zlib.h.
            uncompressed = zlib.decompress(data, -15)
            self._datum_decoder = BinaryDecoder(io.BytesIO(uncompressed))
        elif self.codec == 'snappy':
            import snappy  # Import here b/c we don't expect this to be used.

            # Compressed data includes a 4-byte CRC32 checksum
            length = self.raw_decoder.read_long()
            data = self.raw_decoder.read(length - 4)
            uncompressed = snappy.decompress(data)
            self._datum_decoder = BinaryDecoder(io.BytesIO(uncompressed))
            self.raw_decoder.check_crc32(uncompressed)
        else:
            raise avro.datafile.DataFileException(
                "Unknown codec: %r" % self.codec)


def _make_schema(x, name: str) -> Union[str, dict]:
    assert isinstance(name, str)

    if isinstance(x, numpy.float32):
        return {'name': name, 'type': 'float', 'pytype': 'numpy'}
    if isinstance(x, numpy.float64):
        return {'name': name, 'type': 'double', 'pytype': 'numpy'}
    if isinstance(x, numpy.int32):
        return {'name': name, 'type': 'int', 'pytype': 'numpy'}
    if isinstance(x, numpy.int64):
        return {'name': name, 'type': 'long', 'pytype': 'numpy'}
    if isinstance(x, (numpy.int8, numpy.int16, numpy.uint8, numpy.uint16,
                      numpy.uint32, numpy.uint64)):
        return {
            'name': name,
            'type': 'fixed',
            'size': x.itemsize,
            'pytype': 'numpy',
            'logical_type': x.dtype.name
        }
    if isinstance(x, numpy.ndarray):
        assert len(
            x.shape
        ) == 1, "Multi-dimensional Numpy arrays are not supported. Please convert to a 1-D Numpy array and store it dimensionality info as another datum"
        z = _make_schema(x.dtype.type(), name + '_item')
        return {'name': name, 'type': 'array', 'items': z, 'pytype': 'numpy'}

    if isinstance(x, int):
        return {'name': name, 'type': "int"}
    if isinstance(x, float):
        return {'name': name, 'type': 'double'}
    if isinstance(x, str):
        return {'name': name, 'type': 'string'}
    if isinstance(x, dict):
        fields = []
        for key, val in x.items():
            z = _make_schema(val, key)
            if len(z) < 3:
                fields.append(z)
            else:
                fields.append({'name': key, 'type': z})
        return {'name': name, 'type': 'record', 'fields': fields}
    if isinstance(x, list):
        assert len(
            x
        ) > 0, "empty list is not supported, because its type can not be inferred"
        z0 = _make_schema(x[0], name + '_item')
        if len(x) > 1:
            for v in x[1:]:
                z1 = _make_schema(v, name + '_item')
                assert z1 == z0, \
                    'schema for x[0] ({}): {}; schema for x[?] ({}): {}'.format(
                        x[0], z0, v, z1)
        if len(z0) < 3:
            items = z0['type']
        else:
            items = z0
        return {'name': name, 'type': 'array', 'items': items}

    raise Exception('unrecognized value of type "' + type(x).__name__ + '"')


def make_schema(value, name: str,
                namespace: str) -> str:
    '''
    `value` is a `dict` whose members are either 'simple types' or 'compound types'.

    'simple types' include:
        int, float, str (python types)
        numpy.{int8, int16, int32, int64, uint8, uint16, uint32, uint64, float32, float64}  (numpy types)

    'compound types' include:
        dict: whose elements are simple or compound types
        list: whose elements are all the same simple or compound type
        numpy.ndarray: must be 1-d, with `dtype` being one of the numpy 'simple' type.
    '''
    sch = {'namespace': namespace, **_make_schema(value, name)}
    return json.dumps(sch)


# Using `DataFileWriter` instead of the barebone `DatumWriter`,
# the schema is included in the resultant byte array.
def dump_bytes(value,
               name: str,
               namespace: str) -> bytes:
    schema = make_schema(value, name, namespace)
    buffer = io.BytesIO()
    with DataFileWriter(buffer, schema) as writer:
        writer.append(value)
        writer.flush()
        buffer.seek(0)
        return buffer.getvalue()


def load_bytes(b: bytes, return_schema=False):
    with DataFileReader(io.BytesIO(b)) as reader:
        # We know the stream contains only one record.
        value = list(reader)
        if len(value) == 1:
            value = value[0]
        if return_schema:
            schema = reader.meta['avro.schema'].decode()
            return value, schema
        return value



import json
import pkgutil
import os
import random
import textwrap
import time
from typing import Union, Optional, Iterable, Iterator

import pandas as pd
import requests

from .path import relative_path

# Ideas are borrowed from the python package 'pylivy'.


class SparkSessionError(Exception):
    def __init__(self, name, value, traceback, kind) -> None:
        self._name = name
        self._value = value
        self._tb = traceback
        self._kind = kind
        super().__init__(name, value, traceback)

    def __str__(self) -> str:
        kind = 'Spark' if self._kind == 'spark' else 'PySpark'
        return '{}: {} error while processing submitted code.\nname: {}\nvalue: {}\ntraceback:\n{}'.format(
            self.__module__ + '.' + self.__class__.__name__, kind, self._name,
            self._value, ''.join(self._tb))


def _get_livy_host(livy_server_url: Optional[str] = None):
    '''
    `livy_server_url` is server IP including port number, like '1.2.3.4:8998'.
    '''
    livy_server_url = livy_server_url or os.environ['LIVY_SERVER_URL']

    # remove protocol
    idx = livy_server_url.find('://')
    if idx >= 0:
        livy_server_url = livy_server_url[(idx + 3):]

    return 'http://' + livy_server_url


def unquote(s: str) -> str:
    return s[1:][:-1]  # remove the quotes in the string value


def polling_intervals(start: Iterable[float],
                      rest: float,
                      max_duration: float = None) -> Iterator[float]:
    def _intervals():
        yield from start
        while True:
            yield rest

    cumulative = 0.0
    for interval in _intervals():
        cumulative += interval
        if max_duration is not None and cumulative > max_duration:
            break
        yield interval


class JsonClient:
    def __init__(self, url: str) -> None:
        self.url = url
        self.session = requests.Session()
        self.session.headers.update({'X-Requested-By': 'ambari'})

    def close(self) -> None:
        self.session.close()

    def get(self, endpoint: str = '') -> dict:
        return self._request('GET', endpoint)

    def post(self, endpoint: str, data: dict = None) -> dict:
        return self._request('POST', endpoint, data)

    def delete(self, endpoint: str = '') -> dict:
        return self._request('DELETE', endpoint)

    def _request(self, method: str, endpoint: str, data: dict = None) -> dict:
        url = self.url.rstrip('/') + endpoint
        response = self.session.request(method, url, json=data)
        response.raise_for_status()
        return response.json()


class SparkSession:
    def __init__(self,
                 livy_server_url=None,
                 kind: str = 'spark'):
        '''
        When `kind` is 'spark', this takes Scala code.
        When `kind` is 'pyspark', this takes Python code.
        '''
        if kind == 'scala':
            kind = 'spark'
        assert kind in ('spark', 'pyspark')

        self._host = _get_livy_host(livy_server_url)
        self._client = JsonClient(self._host)
        self._kind = kind
        self._start()

    def _wait(self, endpoint: str, wait_for_state: str):
        intervals = polling_intervals([0.1, 0.2, 0.3, 0.5], 1.0)
        while True:
            rr = self._client.get(endpoint)
            if rr['state'] == wait_for_state:
                return rr
            time.sleep(next(intervals))

    def _start(self):
        r = self._client.post('/sessions', data={'kind': self._kind})
        session_id = r['id']
        self._session_id = session_id
        self._wait('/sessions/{}'.format(session_id), 'idle')

    def run(self, code: str) -> str:
        data = {'code': textwrap.dedent(code)}
        r = self._client.post(
            '/sessions/{}/statements'.format(self._session_id), data=data)
        statement_id = r['id']
        z = self._wait(
            '/sessions/{}/statements/{}'.format(self._session_id,
                                                statement_id), 'available')
        output = z['output']
        if output['status'] == 'error':
            raise SparkSessionError(output['ename'], output['evalue'],
                                    output['traceback'], self._kind)
        assert output['status'] == 'ok'
        return output['data']['text/plain']

    def __del__(self):
        if hasattr(self, '_session_id'):
            try:
                self._client.delete('/sessions/{}'.format(self._session_id))
            finally:
                pass


class ScalaSparkSession(SparkSession):
    def __init__(self, livy_server_url=None):
        super().__init__(livy_server_url, kind='spark')



class PySparkSession(SparkSession):
    def __init__(self, livy_server_url=None):
        super().__init__(livy_server_url, kind='pyspark')
        self._tmp_var_name = '_tmp_' + str(random.randint(100, 100000))
        self.run('import pyspark; import json')

    def run_module(self, module_name: str) -> None:
        '''
        `module_name` is an absolute module name, meaning it's not relative
        to where this code is called. The named module must be reachable
        on the system path via `import`

        Run the source of a Python module in Spark.
        Functions and variables defined in that module are available to subsequent code.
        This module must depend on only Python's stblib and Spark models (plus other modules
        you are sure exist on the Spark cluster). Modules loaded this way should not 'import'
        each other. Imagine the source code of the module is copied and submitted to Spark,
        and the module name is lost.

        The named module is written against the Python version and packages available on the
        Spark cluster. The module is NOT imported by this method; it's entirely possible that
        loading the module in the current environment (not on the Spark cluster) would fail.
        The current method retrieves the source code of the module as text and sends it to Spark.

        Similar concerns about dependencies apply to `run_file`.

        Example (assuming it is meaningful to submit the module `zpz.spark` to Spark):

            sess = PySparkSession()
            sess.run_module('zpz.spark')

        The module shoule 'define' functions, variables only; it should not actually run
        and return anything meaningful.
        '''
        pack = pkgutil.get_loader(module_name)
        self.run_file(pack.path)

    def run_file(self, file_path: str) -> str:
        '''
        `file_path` the absolute path to a Python (script or module) file.

        When you know the relative location of the file, refer to `zpz.path.relative_path`
        and test case `test_file` in `tests/test_spark.py`.

        This returns a string, which is not supposed to be very useful.
        If one wants to capture the result in type-controlled ways, it's better
        to let the file define functions only, then call `fun_file` followed by
        `run_function`.
        '''
        assert file_path.startswith('/')
        code = open(file_path, 'r').read()
        return self.run(code)

    def run_function(self, fname: str, *args, **kwargs):
        '''
        `fname`: name of a function existing in the Spark session environment.

        `args`, `kwargs`: simple arguments to the function, such as string, number, short list.
        '''
        code = f'{fname}(*{args}, **{kwargs})'
        return self.read(code)

    def read(self, code: str):
        '''
        `code` is a variable name or a single, simple expression in the Spark session.        
        '''
        code = '{} = {}; type({}).__name__'.format(self._tmp_var_name, code,
                                                   self._tmp_var_name)
        z_type = unquote(self.run(code))

        if z_type in ('int', 'float', 'bool'):
            z = self.run(self._tmp_var_name)
            return eval(z_type)(z)

        if z_type == 'str':
            z = self.run(self._tmp_var_name)
            return unquote(z)

        if z_type == 'DataFrame':
            code = """\
                for _livy_client_serialised_row in {}.toJSON().collect():
                    print(_livy_client_serialised_row)""".format(
                self._tmp_var_name)
            output = self.run(code)

            rows = []
            for line in output.split('\n'):
                if line:
                    rows.append(json.loads(line))
            z = pd.DataFrame.from_records(rows)
            return z

        try:
            z = self.run('json.dumps({})'.format(self._tmp_var_name))
            zz = json.loads(unquote(z))
            return eval(z_type)(zz)
        except Exception as e:
            raise Exception(
                'failed to fetch an "{}" object: {};  additional error info: {}'.
                format(z_type, z, e))



from collections import OrderedDict

import numpy as np
import pandas as pd


def round_floats(obj, ndigits=2):
    """
    Round floats to specified digits, while
    leaving the data structure and other data types intact.

    The intended use case is for pretty-printing.
    """
    if isinstance(obj, float):
        return round(obj, ndigits)
    if isinstance(obj, dict):
        return dict((k, round_floats(v, ndigits)) for k, v in obj.items())
    if isinstance(obj, (list, tuple)):
        return type(obj)(map(lambda x: round_floats(x, ndigits), obj))
    if isinstance(obj, np.ndarray):
        if issubclass(obj.dtype.type, np.floatting):
            return obj.round(ndigits)
        else:
            return obj
    if isinstance(obj, pd.Series):
        return obj.map(lambda x: round_floats(x, ndigits))
    if isinstance(obj, pd.DataFrame):
        return obj.applymap(lambda x: round_floats(x, ndigits))
    return obj


# Note: take a look at 'serde.py'.
# Might be better option.
def numpy_to_serializable(obj):
    """
    Transform Numpy and pandas components of a data object so that
    the entire object is json-serializable.
    """
    if isinstance(obj, pd.DataFrame):
        index = obj.index.tolist()
        columns = OrderedDict((k, obj[k].values.tolist()) for k in obj)
        return {'pandas.DataFrame': {'index': index, 'columns': columns}}
    if isinstance(obj, pd.Series):
        index = obj.index.tolist()
        values = obj.values.tolist()
        return {'pandas.Series': {'index': index, 'values': values}}
    if isinstance(obj, np.ndarray):
        return {'numpy.ndarray': obj.tolist()}
    if isinstance(obj, (np.integer, np.floating)):
        return {
            'numpy.scalar': {
                'type': obj.dtype.type.__name__,
                'value': np.asscalar(obj)
            }
        }
    if isinstance(obj, dict):
        return type(obj)((k, numpy_to_serializable(v)) for k, v in obj.items())
    if isinstance(obj, (list, tuple)):
        return type(obj)(numpy_to_serializable(v) for v in obj)
    return obj


def numpy_from_serializable(obj):
    """
    Inverse of :func:`numpy_to_serializable`.

    Examples::

        x = 'abc'
        y = np_to_serializable(x)

        z = np_from_serializable(y)
        assert z == x

        yy = json.dumps(y)
        zz = np_from_serializable(json.loads(yy))
        assert zz == x
    """
    if isinstance(obj, dict):
        if len(obj) == 1:
            key = list(obj.keys())[0]
            val = obj[key]
            if key == 'pandas.DataFrame':
                return pd.DataFrame(val['columns'], index=val['index'])
            if key == 'pandas.Series':
                return pd.Series(val['values'], index=val['index'])
            if key == 'numpy.ndarray':
                return np.array(val)
            if key == 'numpy.scalar':
                return eval('np.' + val['type'])(val['value'])
        return {k: numpy_from_serializable(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return type(obj)(numpy_from_serializable(v) for v in obj)
    return obj



from datetime import datetime

from bokeh.plotting import figure
import numpy


def hist(data, bins=20, fill_color="#D95B43", line_color="#033649",
        background_fill_color="#E8DDCB",  **fig_args):
    fig = figure(background_fill_color=background_fill_color, **fig_args)
    hist, edges = numpy.histogram(data, bins=bins)
    fig.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:], 
        fill_color=fill_color, line_color=line_color)
    return fig


def boxplot(data, background_fill_color="#EFE8E2", **fig_args):
    '''
    `data`: a dict of numpy arrays.
    '''
    labels = list(data.keys())
    values = list(data.values())

    q1 = numpy.array([numpy.quantile(v, q=0.25) for v in values])
    q2 = numpy.array([numpy.quantile(v, q=0.5) for v in values])
    q3 = numpy.array([numpy.quantile(v, q=0.75) for v in values])
    iqr = q3 - q1
    upper = q3 + 1.5*iqr
    lower = q1 - 1.5*iqr

    def outliers(x, lower, upper):
        x = x[(x > upper) | (x < lower)]
        return x[~numpy.isnan(x)]

    out = [outliers(x, l, u) for x, l, u in zip(values, lower, upper)]
    out_empty = sum(len(v) for v in out) == 0

    # prepare outlier data for plotting, we need coordinates for every outlier.
    if not out_empty:
        outx = []
        outy = []
        for x, label in zip(out, labels):
            if len(x) > 0:
                outx.extend([label] * len(x))
                outy.extend(x)

    # shrink lengths of stems to be no longer than the minimums or maximums
    qmin = [numpy.min(v) for v in values]
    qmax = [numpy.max(v) for v in values]
    lower = [max(a,b) for a,b in zip(lower, qmin)]
    upper = [min(a,b) for a,b in zip(upper, qmax)]

    x_datetime = isinstance(labels[0], datetime)
    x_unit = 3600000 if x_datetime else 1

    #args = {'background_fill_color': background_fill_color, 'x_range': labels, **figargs}
    args = {'background_fill_color': background_fill_color, **fig_args}
    if x_datetime:
        args['x_axis_type'] = 'datetime'
    else:
        args['x_range'] = labels
    fig = figure(**args)
    fig.xgrid.grid_line_color = None
    fig.ygrid.grid_line_color = "white"
    fig.grid.grid_line_width = 2
    fig.xaxis.major_label_text_font_size= "12pt"

    # stems
    fig.segment(labels, upper, labels, q3, line_color="black")
    fig.segment(labels, lower, labels, q1, line_color="black")

    # boxes
    fig.vbar(labels, 0.7*x_unit, q2, q3, fill_color="#E08E79", line_color="black")
    fig.vbar(labels, 0.7*x_unit, q1, q2, fill_color="#3B8686", line_color="black")

    # whiskers (almost-0 height rects simpler than segments)
    hgt = (numpy.max(upper) - numpy.min(lower)) / 100.
    fig.rect(labels, lower, 0.2*x_unit, hgt, fill_color='black', line_color="black")
    fig.rect(labels, upper, 0.2*x_unit, hgt, fill_color='black', line_color="black")

    # outliers
    if not out_empty:
        fig.circle(outx, outy, size=3, color="#F38630", fill_alpha=0.6)

    return fig


from contextlib import contextmanager

from sqlalchemy.engine import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker


class ManagedScopedSession(object):
    def __init__(self, database, host, port, user, password, **conn_kwargs):
        self._dbsession = scoped_session(
            sessionmaker(
                autocommit=False, autoflush=False, expire_on_commit=False))
        self._database = database

        connection_url = "mysql+pymysql://{user}:{passwd}@{host}:{port}/{database}".format(
            database=database,
            host=host,
            port=port,
            user=user,
            passwd=password)
        for name, value in conn_kwargs.items():
            connection_url += '?{}={}'.format(name, value)

        self._engine = create_engine(connection_url, pool_recycle=3600)
        self._dbsession.configure(bind=self._engine)

    def get_session(self):
        return self._dbsession

    def get_engine(self):
        return self._engine

    @contextmanager
    def session(self):
        try:
            yield self._dbsession
        finally:
            self._dbsession.rollback()
            # self._dbsession.remove()
            # This will rollback un-committed transactions.



import os
import os.path

from distributed import Client as Client


class Dask(Client):
    def __init__(self, scheduler_url=None) -> None:
        scheduler = scheduler_url or os.environ['DASK_SCHEDULER_URL']
        super().__init__(scheduler)


class LocalDask(Client):
    def __init__(self, *args, local_dir=None, **kwargs) -> None:
        local_dir = local_dir or os.path.join(os.environ['DATADIR'], 'dask-worker-space')
        super().__init__(*args, **kwargs, local_dir=local_dir)


import inspect
from pathlib import PurePath


def relative_path(path: str) -> str:
    '''
    Given path `path` relative to the file where this function is called,
    return absolute path.

    For example, suppose this function is called in file '/home/work/src/repo1/scripts/abc.py', then

        relative_path('../../make_data.py')

    returns '/home/work/src/make_data.py', whereas

        relative_path('./details/sum.py')

    returns '/home/work/src/repo1/scripts/details/sum.py'.
    '''
    if path.startswith('/'):
        return path

    caller = inspect.getframeinfo(inspect.stack()[1][0]).filename
    assert caller.endswith('.py')
    p = PurePath(caller).parent
    while True:
        if path.startswith('./'):
            path = path[2:]
        elif path.startswith('../'):
            path = path[3:]
            p = p.parent
        else:
            break
    return str(p.joinpath(path))


import cProfile, pstats
from functools import wraps
from io import StringIO
import time
from typing import Callable
import warnings

import line_profiler


def profiled(top: int = 32, sort_by: str = None,
             prof_file: str = None) -> Callable[[Callable], Callable]:
    """
    Print out profiling info for a test function,
    and optionally dump the profile result in specified file for later inspection.

    Args:
        top: number of top items to print.
        sort_by: list of sort keys for print-out, if not `None`.
            If `None` (the default), a default list of sort keys are used.
            If `[]`, no print-out.
        prof_file: file name for a dump of the profile.
            Use `snakeviz` to view the content of this file later.
            If `None`, a default file name is used.
            If `''`, profile is not saved.

    Usage::

        @profiled()
        def func():
            ...

        @profiled(top=20)
        def func():
            ...

    In a test function, ::

        def test_abc():
            @profiled()
            def do_it():
                ...

            do_it()

    """

    if sort_by is None:
        sort_by = ['cumtime', 'tottime']

    if isinstance(sort_by, str):
        sort_by = [sort_by]

    if prof_file is None:
        prof_file = 'cprofile.out'

    def mydecorator(func):
        @wraps(func)
        def profiled_func(*args, **kwargs):
            profile = cProfile.Profile()
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()

            for sb in sort_by:
                s = StringIO()
                pstats.Stats(profile, stream=s).sort_stats(sb).print_stats(top)
                print('')
                print(s.getvalue())

            if prof_file:
                profile.dump_stats(prof_file)
                print('')
                print('profiling results are saved in', prof_file,
                      '; view its content using `snakeviz`')

            return result

        return profiled_func

    return mydecorator


def lineprofiled(*funcs) -> Callable[[Callable], Callable]:
    """
    A line-profiling decorator.

    Args:
        funcs: functions (function objects, not function names) to be line-profiled.
            If no function is specified, the function being decorated is profiled.

    Example:

            @lineprofiled()
            def myfunc(a, b):
                g(a)
                f(b)

            @lineprofiled(g, f)
            def yourfunc(a, b):
                x = g(a)
                y = f(b)
                s(x, y)
    """

    def mydecorator(func):
        nonlocal funcs
        if not funcs:
            funcs = [func]

        @wraps(func)
        def profiled_func(*args, **kwargs):
            func_names = [f.__name__ for f in funcs]
            profile = line_profiler.LineProfiler(*funcs)
            z = profile.runcall(func, *args, **kwargs)

            profile.print_stats()
            stats = profile.get_stats()
            if not stats.timings:
                warnings.warn("No profile stats.")
                return z

            for key, timings in stats.timings.items():
                if key[-1] in func_names:
                    if len(timings) > 0:
                        func_names.remove(key[-1])
                        if not func_names:
                            break
            if func_names:
                # Force warnings.warn() to omit the source code line in the message
                formatwarning_orig = warnings.formatwarning
                warnings.formatwarning = lambda message, category, filename, lineno, line=None: \
                    formatwarning_orig(message, category, filename, lineno, line='')
                warnings.warn("No profile stats for %s." % str(func_names))
                # Restore warning formatting.
                warnings.formatwarning = formatwarning_orig

            return z

        return profiled_func

    return mydecorator


def timed(func: Callable) -> Callable:
    @wraps(func)
    def profiled_func(*args, **kwargs):
        time0 = time.perf_counter()
        result = func(*args, **kwargs)
        time1 = time.perf_counter()

        print('')
        print('Function `', func.__name__, '` took ', time1 - time0,
              'seconds to finish')
        print('')
        return result

    return profiled_func


class Timer:
    def __init__(self):
        self._running = False

    def start(self):
        self._t_start = time.perf_counter()
        self._t_stop = None
        self._running = True
        return self

    def stop(self):
        if self._running:
            self._t_stop = time.perf_counter()
            self._running = False
        return self

    @property
    def seconds(self):
        if self._running:
            return time.perf_counter() - self._t_start
        return self._t_stop - self._t_start

    @property
    def milliseconds(self):
        return self.seconds * 1000



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

full = '0'
exec((HERE / 'src/zpz/version.py').read_text(), globals())
version = full
del full

setup(
    name='zpz',
    version=version,
    description='Various Python utilities',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/zpz/utilities.py',
    license='BSD',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    include_package_data=True,
)



import configparser
import os
import textwrap
from typing import Sequence

from .sql import pretty_sql


class ColumnMapper(object):
    """
    Definition of mapping from columns in a source table to a single column in a target table.
    """

    def __init__(self, name, type, select=None):
        """
        Args:
            name (str): name of the resultant column.
            type (str): SQL data type of the resultant column.
            select (str): SQL expression that defines the resultant column.
                The expression typically involves columns (of a source table),
                constants, SQL built-in functions, and operators.
                If ``None``, the resultant column is a column from the source table
                with the same name.

        Examples:
            ::

                ('def', 'double')

            'def' in the original table is carried over to the new table with no change.

            ::

                ('abc', 'double', 'col1 * EXP(col2)')

            columns 'col1' and 'col2' are combined to create 'abc' in the new table.

            ::

                ('pg', 'bigint', 'st')

            'st' in the original table is renamed to 'pg' in the new table.
        """
        self.name = name
        self.type = type.upper()

        self.define = '{} {}'.format(self.name, self.type)
        """Used in table creation statements to define the column."""

        if select:
            self.select = '{} AS {}'.format(select, name)
        else:
            self.select = name
        """Used in ``SELECT`` statements to extract the column."""

    def __repr__(self):
        select = self.select.split()[0]
        if select == self.name:
            select = None
        if select:
            return "{}('{}', '{}', '{}')".format(self.__class__.__name__,
                                                 self.name, self.type, select)
        else:
            return "{}('{}', '{}')".format(self.__class__.__name__, self.name,
                                           self.type)


class RowMapper(object):
    """Definition of mapping from a row in a source table to a row in a target table.

    A ``RowMapper`` is a list of :class:`ColumnMapper` s.
    """

    def __init__(self, *column_mappers):
        """
        Args:
            column_mappers: ``ColumnMapper`` objects or iterables
                that would serve as input to the initiator of
                :class:`ColumnMapper`.

                Examples::

                    ColumnMapper('def', 'double'), ColumnMapper('abc', 'double', 'col1 * EXP(col2)')
                    ('def', 'double'), ColumnMapper('abc', 'double', 'col1 * EXP(col2)')
                    ('def', 'double'), ('abc', 'double', 'col1 * EXP(col2)')

        """
        if column_mappers:
            self._col_mappers = [
                v if isinstance(v, ColumnMapper) else ColumnMapper(*v)
                for v in column_mappers
            ]
        else:
            self._col_mappers = []

    def add_cols(self, *column_mappers):
        self._col_mappers.extend(
            v if isinstance(v, ColumnMapper) else ColumnMapper(*v)
            for v in column_mappers)
        return self

    @property
    def selects(self):
        return [x.select for x in self._col_mappers]

    @property
    def defines(self):
        return [x.define for x in self._col_mappers]

    @property
    def names(self):
        return [x.name for x in self._col_mappers]

    def map(self, table_or_sql_in):
        sql = """\
        SELECT {}
        FROM {} AS t_in
        """.format(
            ',\n'.join(self.selects),
            table_or_sql_in.as_subclause,
        )
        return pretty_sql(sql)


def select_to_table(sql_select, table_out, col_defs=None):
    """Create a table that is populated by the output of a ``SELECT`` statement.

    Args:
        sql_select (str): a ``SELECT`` statement.
        table_out (str): name of output table in the format of 'database.table'.
        col_defs (str): definition of columns, such as::

            'abc INT, efg CHAR(20)'

    Returns:
        str: a SQL statement for creating and populating the table.
    """
    if col_defs:
        sql = """\
        DROP TABLE IF EXISTS {t_out}
        ;
        CREATE TABLE {t_out} (
            {defs}
            )
            STORED AS PARQUET
        ;
        INSERT OVERWRITE TABLE {t_out}
        {select}\
        """.format(
            t_out=table_out, defs=col_defs, select=sql_select)
    else:
        sql = """\
        DROP TABLE IF EXISTS {t_out}
        ;
        CREATE TABLE {t_out}
            STORED AS PARQUET
        AS
        {select}\
        """.format(
            t_out=table_out, select=sql_select)

    return pretty_sql(sql)



import logging
import os
from pprint import pprint
from typing import Sequence, Tuple, List, Dict, Union

from impala import dbapi
from retrying import retry

from .sql import split_sql, SQLClient
from ..s3 import Bucket

logging.getLogger('impala.hiveserver2').setLevel(logging.WARNING)
logging.getLogger('impala._thrift_api').setLevel(logging.WARNING)

logger = logging.getLogger(__name__)


class Hive(SQLClient):
    def __init__(self,
                 *,
                 user,
                 host,
                 port=10000,
                 auth_mechanism='PLAIN',
                 exec_engine='tez',
                 queue: str=None,
                 dynamic_partition: bool=True,
                 configuration: dict=None):
        super().__init__(
            conn_func=dbapi.connect,
            cursor_args={'user': user},
            host=host,
            port=int(port),
            user=user,
            auth_mechanism=auth_mechanism)
        config = configuration or {}
        if exec_engine == 'tez':
            config['hive.execution.engine'] = 'tez'
            if queue:
                config['tez.queue.name'] = queue
        if dynamic_partition:
            config['hive.exec.dynamic.partition'] = 'true'
            config['hive.exec.dynamic.partition.mode'] = 'nonstrict'

        config['hive.optimize.s3.query'] = 'true'
        config['hive.enforce.bucketing'] = 'true'

        self._configuration = config

    def _parse_sql(self, sql: str) -> Tuple[List[str], Dict[str, str]]:
        assert isinstance(sql, str)
        sql = split_sql(sql)
        configuration = {}
        ss = []
        for s in sql:
            if s.strip().upper().startswith('SET '):
                a, b = s[4:].split('=')
                configuration[a.strip()] = b.strip()
            else:
                ss.append(s)
        return ss, configuration

    def read(self, sql: str):
        sqls, config = self._parse_sql(sql)
        assert len(sqls) == 1
        sql = sqls[0]
        config = {**self._configuration, **config}
        return super().read(sql, configuration=config)

    def write(self, sql: str):
        sqls, config = self._parse_sql(sql)
        config = {**self._configuration, **config}
        return super().write(sqls, configuration=config)

    def get_databases(self) -> Sequence[str]:
        self._cursor.get_databases()
        return [v[0] for v in self._cursor.fetchall()]

    def has_database(self, db_name: str) -> bool:
        return self._cursor.database_exists(db_name)

    def get_tables(self, db_name: str) -> Sequence[str]:
        self._cursor.get_tables(db_name)
        return [v[2] for v in self._cursor.fetchall()]

    def has_table(self, db_name: str, tb_name: str) -> bool:
        return self._cursor.table_exists(tb_name, db_name)

    def get_table_schema(self, db_name: str,
                         tb_name: str) -> Sequence[Tuple[str, str]]:
        return self._cursor.get_table_schema(tb_name, db_name)

    def show_create_table(self, db_name: str, tb_name: str) -> None:
        sql = f'SHOW CREATE TABLE {db_name}.{tb_name}'
        z = self.read(sql).fetchall()
        z = '\n'.join(v[0] for v in z)
        print(z)

    def describe_table(self, db_name: str, tb_name: str) -> None:
        sql = f'DESCRIBE FORMATTED {db_name}.{tb_name}'
        z = self.read(sql).fetchall_pandas()
        print(z)


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000, stop_max_attempt_number=7)
def get_hive(user, host, queue=None, **kwargs):
    # `host`: hive server url.
    return Hive(user=user, queue=queue, host=host, **kwargs)


class HiveTableMixin:
    def __init__(self, *,
            db_name: str,
            tb_name: str,
            columns: List[Tuple[str, str]],
            partitions: List[Tuple[str,str]]=None,
            stored_as: str='ORC',
            field_delimiter: str='\\t',
            compression: str='ZLIB',
            location: str=None) -> None:
        self.db_name = db_name
        self.tb_name = tb_name
        self.columns = [(name, type_.upper()) for (name, type_) in columns]
        if partitions:
            self.partitions = [(name, type_.upper()) for (name, type_) in partitions]
        else:
            self.partitions = []

        stored_as = stored_as.upper()
        assert stored_as in ('ORC', 'PARQUET', 'TEXTFILE')
        self.stored_as = stored_as
        self.field_delimiter = field_delimiter
        self.compression = compression

        self.external = False
        self.s3external = False
        if location:
            self.location = location.rstrip('/') + '/'   # Ensure trailing '/'
            loc = self.location
            if loc.startswith('s3://') or loc.startswith('s3n://'):
                if loc.startswith('s3://'):
                    z = self.location[len('s3://') :]
                else:
                    z = self.location[len('s3n://') :]
                assert '/' in z
                self._s3_bucket_key = z[: z.find('/')]
                self._s3_bucket_path = z[(z.find('/') + 1) :]
                self.s3external = True
            self.external = True

    @property
    def full_name(self):
        return self.db_name + '.' + self.tb_name

    def create(self, engine, drop_if_exists: bool=False) -> None:
        '''
        `engine` is a `Hive` or `Athena` object.
        '''
        def collapse(spec):
            return ', '.join(name + ' ' + type_ for (name, type_) in spec)

        columns = collapse(self.columns)

        if self.partitions:
            partitions = f"PARTITIONED BY ({collapse(self.partitions)})"
        else:
            partitions = ''

        if self.external:
            location = f"LOCATION '{self.location}'"
            external = 'EXTERNAL'
        else:
            location = ''
            external = ''

        if self.stored_as in ('ORC', 'PARQUET'):
            stored_as = f'''
                STORED AS {self.stored_as} 
                {location}
                TBLPROPERTIES ('{self.stored_as.lower()}.compress' = '{self.compression}')
                '''
        else:
            stored_as = f'''
                ROW FORMAT DELIMITED FIELDS TERMINATED BY '{self.field_delimiter}'
                STORED AS {self.stored_as}
                {location}
                '''

        sql = f'''
            CREATE {external} TABLE {self.full_name}
            ({columns})
            {partitions}
            {stored_as}
            '''

        if drop_if_exists:
            self.drop(engine)

        engine.write(sql)
        if self.external:
            engine.write(f'MSCK REPAIR TABLE {self.full_name}')

    def drop(self, engine) -> None:
        engine.write(f'DROP TABLE IF EXISTS {self.full_name}')

    def _partition_values_to_path(self, *partition_values) -> str:
        assert 0 < len(partition_values) <= len(self.partitions)
        return '/'.join(f"{k}={v}" for v, (k, _) in zip(partition_values, self.partitions[: len(partition_values)]))

    def _partition_values_to_condition(self, *partition_values) -> str:
        '''
        `partition_values` appear in the order they are listed
        in the table definition.
        '''
        assert 0 < len(partition_values) <= len(self.partitions)
        elements = []
        for v, (k, t) in zip(partition_values, self.partitions[: len(partition_values)]):
            if t.lower() in ('string', 'char', 'varchar'):
                elements.append(f"{k}='{v}'")
            else:
                elements.append(f"{k}={v}")
        return ', '.join(elements)

    def get_partitions(self, engine, *partition_values) -> List[str]:
        if not self.partitions:
            raise Exception('not a partitioned table')
        sql = f'SHOW PARTITIONS {self.full_name}'
        if partition_values:
            cond_str = self._partition_values_to_condition(*partition_values)
            sql += f' PARTITION({cond_str})'
        z = engine.read(sql).fetchall()
        return [v[0] for v in z]

    def show_partitions(self, engine, *partition_values) -> None:
        pprint(sorted(self.get_partitions(engine, *partition_values)))

    def show_partition_counts(self, engine, *partition_values):
        if not self.partitions:
            raise Exception('not a partitioned table')
        cols = ', '.join(v[0] for v in self.partitions)
        sql = f'''
            SELECT
                {cols},
                COUNT(*) AS count
            FROM {self.full_name}'''
        if partition_values:
            cond_str = self._partition_values_to_condition(*partition_values)
            sql = sql + f'''
                WHERE {cond_str}'''
        sql = sql + f'''
            GROUP BY {cols}
            ORDER BY {cols}'''
        z = engine.read(sql).fetchall_pandas()
        print(z)

    def drop_partitions(self, engine, *partition_values):
        if partition_values:
            cond_str = self._partition_values_to_condition(*partition_values)
            sql = f"ALTER TABLE {self.full_name} DROP IF EXISTS PARTITION({cond_str})"
            engine.write(sql)
        else:
            logger.warning(f'dropping all partitions in table {self.full_name}')
            parts = self.get_partitions(engine)
            for p in parts:
                cond_str = p.split('/')[0]
                sql = f"ALTER TABLE {self.full_name} DROP IF EXISTS PARTITION({cond_str})"
                engine.write(sql)

    def add_partitions(self, engine, *partition_values):
        '''
        Add partitions for an external table to the table meta data.

        Use case is adding partition to the meta data of an external table
        after new data has been added w/o the knowledge of the meta store.

        `*partition_values`: values of first partition columns; update below that.

        If empty, update under root location.
        '''
        assert self.s3external

        existing_partitions = self.get_partitions(engine, *partition_values)
        if not partition_values:
            root = self.location
        else:
            parts_path_str = self._partition_values_to_path(*partition_values)
            parts_cond_str = self._partition_values_to_condition(*partition_values)
            root = self.location + parts_path_str + '/'

        bucket = Bucket(self._s3_bucket_key)
        files = bucket.ls(root, recursive=True)

        if len(partition_values) == len(self.partitions):
            if list(files):
                sql = f'''ALTER TABLE {self.full_name} ADD PARTITION({parts_cond_str}) LOCATION '{self.location + parts_path_str}' '''
                logger.debug(sql)
                engine.write(sql)
            return

        dirs = set(v[: v.rfind('/')] for v in files if '/' in v)  # no trailing '/'
        for dd in dirs:
            path_str = dd
            if partition_values:
                path_str = parts_path_str + '/' + path_str
            if path_str not in existing_partitions:
                cond_str = ', '.join(f"{k}='{v}'" for k,v in (v.split('=') for v in dd.split('/')))
                if partition_values:
                    cond_str = parts_cond_str + ', ' + cond_str
                sql = f'''ALTER TABLE {self.full_name} ADD PARTITION({cond_str}) LOCATION '{self.location + path_str}' '''
                logger.debug(sql)
                engine.write(sql)

    def update_partitions(self, engine, *partition_values):
        assert self.s3external
        self.drop_partitions(engine, *partition_values)
        self.add_partitions(engine, *partition_values)

    def partition_location(self, *partition_values):
        if not self.s3external:
            raise NotImplementedError
        if not partition_values:
            return self.location
        path = self.location + self._partition_values_to_path(*partition_values) + '/'
        return path

    def purge_data(self, *partition_values) -> int:
        if not self.s3external:
            raise NotImplementedError
        bucket = Bucket(self._s3_bucket_key)
        path = self.partition_location(*partition_values)
        return bucket.delete_tree(path)


class HiveTable(HiveTableMixin):
    def __init__(self, location: str=None, **kwargs) -> None:
        if location:
            assert location.startswith('s3n://')
        super().__init__(**kwargs, location=location)

    @classmethod
    def from_athena_table(cls, table: 'AthenaTable', db_name: str, tb_name: str=None) -> 'HiveTable':
        '''
        Use case is analogous to `athena.AthenaTable.from_hive_table`.
        '''
        assert table.s3external
        location = table.location
        if location.startswith('s3://'):
            location = 's3n://' + location[len('s3://') :]
        return cls(
            db_name=db_name,
            tb_name=tb_name or table.tb_name,
            columns=table.columns,
            partitions=table.partitions,
            stored_as=table.stored_as,
            field_delimiter=table.field_delimiter,
            compression=table.compression,
            location=location)



"""
A commonly used config file format is ``INI``. Here is an example file::

    [server]
    host = 1.2.3.4
    port = 28

    # This block concerns our secrete sauce
    speed =
    # fill in an integer greater than 1000000 above
    use-power
    ; leave this alone

    nocomment = this; that
    semicolon = ;

    [client settings]
    browser = chrome only

A few points about the format:

1. Feel free to use blank lines for content organization, even within a single 'section'.
2. A comment line starts with either `#` or `;`, *with no leading space*.
3. 'speed' have value '' (empty string).
4. 'use-power' has value of Python's ``None``.
5. Section name can contain space.
6. 'true', 'True', 'false', 'False', 'yes', 'Yes', 'no', 'No' are recognized in expected ways.

In light of the peculiarities of the syntax for comments, we recommend the following regarding comments:

- Do not use in-line comments.
- Use whole-line comments starting with '#', with no leading space.
- Use blank lines to help organization.

**Caution**: because ``configparser.ConfigParser`` ignores comments,
if you read in a config file using ``configparser.ConfigParser``, make changes
to the object, then save the changed object its ``write`` method,
comments in the original file will be lost.

Do not quote the values---quotation marks will be literal parts of the value.

For complex values, use JSON strings. Note, quotation marks in JSON strings should be *double quotes*.
"""

import configparser
import os


def read_ini_config(file_name: str,
                    section_name: str = None,
                    allow_no_value=True) -> dict:
    """
    Get the content of an INI-style config file (for read-only).

    Args:
        file_name: config file name including path so that the file can be found
            by this function; an absolute full path will be most reliable.
        section_name: name of section in the INI file; if `None`, whole file is returned.

    The returned object works much like a `dict` thanks to the
    *mapping protocol access* capabilities in Python 3.
    """
    conf = configparser.ConfigParser(allow_no_value=allow_no_value)
    conf.read(file_name)
    if section_name:
        return conf[section_name]
    return conf



"""
Configure logging, mainly the format.

A call to the function ``config_logger`` in a launching script is all that is needed to set up the logging format.
Usually the 'level' argument is the only argument one needs to customize::

  config_logger(level='info')

If `level` is not specified, environment variable `LOGLEVEL` is used;
if that is not set, a default level (currently 'info') is used.

Do not call this in library modules.
Library modules should have ::

   logger = logging.getLogger(__name__)

and then just use ``logger`` to write logs without concern about formatting,
destination of the log message, etc.
"""
from datetime import datetime
import logging
from logging import Formatter
import time
from typing import Union

import pytz

#import os
# raiseExceptions = os.environ.get('ENVIRONMENT_TYPE', None) in ('test', 'dev')
# When exceptions are raised during logging, then,
# The default implementation of handleError() in Handler checks to see if a module-level variable,
# raiseExceptions, is set.
# If set, a traceback is printed to sys.stderr.
# If not set, the exception is swallowed.
#
# If no logging configuration is provided, then for Python 2.x,
#   If logging.raiseExceptions is False (production mode), the event is silently dropped.
#   If logging.raiseExceptions is True (development mode), a message 'No handlers could be found for logger X.Y.Z' is printed once.

logging.logThreads = 0
logging.logProcesses = 0


def _make_config(
        *,
        level: Union[str, int]='info',
        format: str='[%(asctime)s; %(name)s, %(funcName)s, %(lineno)d; %(levelname)s]    %(message)s',
        timezone: str='US/Pacific',
        datefmt: str='%Y-%m-%d %H:%M:%S %Z',
        **kwargs) -> dict:
    # 'level' is string form of the logging levels: 'debug', 'info', 'warning', 'error', 'critical'.
    if level not in (logging.DEBUG, logging.INFO, logging.WARNING,
                     logging.ERROR, logging.CRITICAL):
        level = getattr(logging, level.upper())

    if timezone.lower() == 'UTC':
        Formatter.converter = time.gmtime
    elif timezone.lower() == 'local':
        Formatter.converter = time.localtime
    else:
        def custom_time(*args):
            utc_dt = pytz.utc.localize(datetime.utcnow())
            my_tz = pytz.timezone(timezone)
            converted = utc_dt.astimezone(my_tz)
            return converted.timetuple()
        Formatter.converter = custom_time

    return dict(format=format, datefmt=datefmt, level=level, **kwargs)


def config_logger(**kwargs) -> None:
    logging.basicConfig(**_make_config(**kwargs))



from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.encoders import encode_base64
import smtplib
import os
import threading


def _generate_msg(*, send_from, send_to, subject, text, files=None):
    """
    Args:
        send_to: comma-separated email addresses.
        files: either a list of full-path file names,
            or a list of tuples, each tuple containing
            file name (no path; intended to be meaningful to the email receiving party)
            and the file content as a binary blob.
    """
    msg = MIMEMultipart()
    msg['From'] = send_from
    msg['Reply-to'] = send_from
    msg['To'] = send_to
    msg['Subject'] = subject
    msg.attach(MIMEText(text, 'plain'))

    if files is None:
        files = []
    assert isinstance(files, list)
    if files is None:
        files = []
    for f in files:
        part = MIMEBase('application', "octet-stream")
        if isinstance(f, str):
            part.set_payload(open(f, "rb").read())
            encode_base64(part)
            part.add_header('Content-Disposition',
                            'attachment; filename="%s"' % os.path.basename(f))
        else:
            assert isinstance(f, tuple)
            part.set_payload(f[1])
            encode_base64(part)
            part.add_header('Content-Disposition',
                            'attachment; filename="{}"'.format(f[0]))

        msg.attach(part)

    return msg


def _send_mail(server, msg):
    server.sendmail(msg['From'], msg['To'].split(','), msg.as_string())


class Mailer(object):
    def __init__(self, *, host, port, account, passwd):
        """
        Args:
            host: SMTP mail server.
            port: port number; an `int`, but can be a `str` (will be converted in the function).
            account: email account on the server.
            passwd: password for `account` on `host`.

        If you use a Gmail account to send emails, use
        'smtp.gmail.com' for `host` and 587 for `port`.
        """
        self._mail_server_args = dict(
            host=host, port=int(port), account=account, passwd=passwd)

    def _connect(self):
        """
        Todo:
            Should we establish connection upon object initiation?
            For how long will the connection be active?
        """
        # try:
        #     if 'accepting connections' in sh.Command('sudo')('/etc/init.d/sendmail', 'status', __tty_out=False):
        #         server = smtplib.SMTP('localhost', timeout=timeout)
        #         server.ehlo_or_helo_if_needed()
        #         return server
        #     raise Exception('no SMTP server on localhost')
        # except Exception as e:
        # Had trouble using 'sudo' with password passing in.
        server = smtplib.SMTP(self._mail_server_args['host'],
                              self._mail_server_args['port'])
        #server.ehlo()
        server.starttls()
        #server.ehlo()
        server.login(self._mail_server_args['account'],
                     self._mail_server_args['passwd'])
        return server

    def send(self, *, send_from, send_to, subject, text, files=None):
        """
        Args:
            send_from: an email address.
            send_to: a string of comma-separated email addresses.
            files (list): names of files to attach, with full path.

        If Gmail server is used, the 'From' field of the actually sent email will be
        the email address specified by ``self._mail_server_args['account']``,
        and not the ``send_from`` here,
        but we're filling 'Reply-to' with the value of ``send_from``.

        A common use case is to have fixed ``send_from`` and ``send_to`` but send multiple emails.
        In this case, do something like the following::

            from functools import partial

            mailer = Mailer(**server_args)
            my_mailer = partial(mailer.send, send_from='abc@gmail.com', send_to='def@myjob.com')

            my_mailer(subject='abc', text='test')
            my_mailer(subject='ERROR', text='see what you did!')

            # And pass the object 'my_mailer' around to send emails from within other code blocks.
        """
        msg = _generate_msg(
            send_from=send_from,
            send_to=send_to,
            subject=subject,
            text=text,
            files=files)
        server = self._connect()
        thr = threading.Thread(
            target=_send_mail,
            kwargs={
                'server': server,
                'msg': msg
            },
        )
        thr.start()

        # TODO: take care of `server.quit()`.
